// ___SETA43  _____
// 19/04/2025 _____
// RADIO Si4732 FM-AM
// Arduino 1.8.19
// seta43.duckdns.org
// seta43.blogspot.com

/*
  1) FM, AM (MW and SW) and SSB (LSB and USB);
  2) Audio bandwidth filter 0.5, 1, 1.2, 2.2, 3 and 4kHz;
  3) 22 commercial and ham radio bands pre configured;
  4) BFO Control; and
  5) Frequency step switch (1, 5 and 10kHz);
  
  ATTENTION: 
  SI473X bases device  and Arduino Atmega328 based device wire up. 

  | Device name               | Device Pin / Description  |  Arduino Pin  |
  | ----------------          | --------------------      | ------------  |
  | Display OLED              |                           |               |
  |                           | SDA                       |     A4        |
  |                           | CLK                       |     A5        |
  |     Si4735                |                           |               |
  |                           | RESET (pin 15)            |     12        |
  |                           | SDIO (pin 18)             |     A4        |
  |                           | SCLK (pin 17)             |     A5        |
  |     Buttons               |                           |               |
  |                           | MENU                      |      4        |
  |    Encoder                |                           |               |
  |                           | A                         |      2        |
  |                           | B                         |      3        |

  By Ricardo Lima Caratti, Nov 2019.
*/ 
#define FRECUSTART 9470

#include <SI4735.h>
#include <Tiny4kOLED.h>

#include <patch_init.h> // SSB patch for whole SSBRX initialization string

const uint16_t size_content = sizeof ssb_patch_content; // see ssb_patch_content in patch_full.h or patch_init.h

#define FM_BAND_TYPE 0
#define MW_BAND_TYPE 1
#define SW_BAND_TYPE 2
#define LW_BAND_TYPE 3

// OLED Diaplay constants
#define I2C_ADDRESS 0x3C

#define RESET_PIN 12

// Buttons controllers
#include "Rotary.h";
#define ROTARY_PIN1 3
#define ROTARY_PIN2 2
#define BUTTON_PIN  4
#define CLICKS_PER_STEP   2   // this number depends on your rotary encoder 

Rotary r;
int rDir;

#define MIN_ELAPSED_TIME 100
#define MIN_ELAPSED_RSSI_TIME 150

#define DEFAULT_VOLUME 45 // change it for your favorite sound volume

#define FM 0
#define LSB 1
#define USB 2
#define AM 3
#define LW 4

#define SSB 1

const char *bandModeDesc[] = {"FM ", "LSB", "USB", "AM "};
char currentMode = FM;

bool bfoOn = false;
bool disableAgc = 1;
bool ssbLoaded = false;
bool fmStereo = true;

int currentBFO = 0;

//long elapsedRSSI = millis();
//long elapsedButton = millis();

// Encoder control variables
volatile int encoderCount = 0;

// Some variables to check the SI4735 status
uint16_t currentFrequency;
uint16_t previousFrequency;
uint8_t currentStep = 1;
uint8_t currentBFOStep = 25;

uint8_t seekDirection = 1;

uint8_t bwIdxSSB = 2;
const char *bandwidthSSB[] = {"1.2", "2.2", "3.0", "4.0", "0.5", "1.0"};

char bwIdxAM = 1;
const char *bandwidthAM[] = {"6", "4", "3", "2", "1", "1.8", "2.5"};

/*
   Band data structure
*/
typedef struct
{
  uint8_t bandType;     // Band type (FM, MW or SW)
  uint16_t minimumFreq; // Minimum frequency of the band
  uint16_t maximumFreq; // maximum frequency of the band
  uint16_t currentFreq; // Default frequency or current frequency
  uint16_t currentStep; // Defeult step (increment and decrement)
} Band;

/*
   Band table
*/
Band band[] = {
  {FM_BAND_TYPE, 8400, 10800, 10350, 10},
  {LW_BAND_TYPE, 100, 510, 300, 1},
  {MW_BAND_TYPE, 520, 1720, 810, 10},
  {SW_BAND_TYPE, 1800, 3500, 1900, 1}, // 160 meters
  {SW_BAND_TYPE, 3500, 4500, 3700, 1}, // 80 meters
  {SW_BAND_TYPE, 4500, 5500, 4850, 5},
  {SW_BAND_TYPE, 5600, 6300, 6000, 5},
  {SW_BAND_TYPE, 6800, 7800, 7200, 5}, // 40 meters
  {SW_BAND_TYPE, 9200, 10000, 9600, 5},
  {SW_BAND_TYPE, 10000, 11000, 10100, 1}, // 30 meters
  {SW_BAND_TYPE, 11200, 12500, 11940, 5},
  {SW_BAND_TYPE, 13400, 13900, 13600, 5},
  {SW_BAND_TYPE, 14000, 14500, 14200, 1}, // 20 meters
  {SW_BAND_TYPE, 15000, 15900, 15300, 5},
  {SW_BAND_TYPE, 17200, 17900, 17600, 5},
  {SW_BAND_TYPE, 18000, 18300, 18100, 1},  // 17 meters
  {SW_BAND_TYPE, 21000, 21900, 21200, 1},  // 15 mters
  {SW_BAND_TYPE, 24890, 26200, 24940, 1},  // 12 meters
  {SW_BAND_TYPE, 26200, 27900, 27500, 1},  // CB band (11 meters)
  {SW_BAND_TYPE, 28000, 30000, 28400, 1}
}; // 10 meters

const int lastBand = (sizeof band / sizeof(Band)) - 1;
char bandIdx = 0;

uint8_t rssi = 0;
uint8_t stereo = 1;
char volume = DEFAULT_VOLUME;

SI4735 si4735;


///////////////////////////////////
static char *bandName[]= {
"FM   ",
"LW     ",
"MW     ",
"SW 1.8M  ",
"SW 3.5M  ",
"SW 4.5M  ",
"SW 5.6M  ",
"SW 6.8M  ",
"SW 9.2M  ",
"SW 10 M  ",
"SW 11.2M  ",
"SW 13.4M  ",
"SW 14M  ",
"SW 15M  ",
"SW 17.2M  ",
"SW 18M  ",
"SW 21M  ",
"SW 24.8M  ",
"SW 26.2M  ",
"SW 28M  "
};

static char *menus[]={"EXIT","Volumen","Banda","Decode","AGC","BwAM","BwSSB","STEP","BFO","NoInfo"};
#define NMENUS (sizeof(menus)/sizeof(menus[0]))

static char *onOff[]={"OFF ","ON "};

int ti;
char f2;
char info=1;
char modoMenu=0;
char cadena[40];


void setup()
{
  Serial.begin(9600);
  while (!Serial) {;}
  
  // Encoder pins
  r.begin(ROTARY_PIN1, ROTARY_PIN2, CLICKS_PER_STEP);
  r.setLeftRotationHandler(showDirection);
  r.setRightRotationHandler(showDirection);
  pinMode(ROTARY_PIN1, INPUT_PULLUP); 
  pinMode(ROTARY_PIN2, INPUT_PULLUP);
  pinMode(BUTTON_PIN, INPUT_PULLUP); 

  oled.begin();
  oled.clear();
  oled.on();
  oled.setFont(FONT6X8);
  
  // Splash - Change it for your introduction text.
  oled.setCursor(40, 0);
  oled.print(F("SI4735"));
  oled.setCursor(20, 1);
  oled.print(F("Arduino Library"));
  delay(5);
  oled.setCursor(15, 2);
  oled.print(F("All in One Radio"));
  delay(5);
  oled.setCursor(10, 3);
  oled.print(F("V1.1.5 - By PU2CLR"));
  delay(500);
  oled.clear();
  oled.setCursor(10, 3);
  oled.print(F("Modif. SETA43"));
  // end Splash

  si4735.getDeviceI2CAddress(RESET_PIN); // Looks for the I2C bus address and set it.  Returns 0 if error
  si4735.setup(RESET_PIN, FM_BAND_TYPE);

  delay(500);  
  // Set up the radio for the current band (see index table variable bandIdx )
  useBand();

  currentFrequency  = si4735.getFrequency();
  previousFrequency=0;

  currentFrequency  = FRECUSTART;
  delay(500);
  si4735.setFrequency(FRECUSTART);
  previousFrequency=0;
  si4735.setVolume(volume);
  oled.clear();
  showStatus();
}


// Show current frequency

void showFrequency()
{
  String freqDisplay;
  String unit;
  String bandMode;
  int divider = 1;
  int decimals = 3;
  if (band[bandIdx].bandType == FM_BAND_TYPE)
  {
    divider = 100;
    decimals = 1;
    unit = "MHz";
  }
  else if (band[bandIdx].bandType == MW_BAND_TYPE || band[bandIdx].bandType == LW_BAND_TYPE)
  {
    divider = 1;
    decimals = 0;
    unit = "kHz";
  }
  else
  {
    divider = 1000;
    decimals = 3;
    unit = "kHz";
  }

  if ( !bfoOn )
    freqDisplay = String((float)currentFrequency / divider, decimals);
  else
    freqDisplay = ">" + String((float)currentFrequency / divider, decimals) + "<";

  oled.setCursor(38, 0);
  oled.print("        ");
  oled.setCursor(38, 0);
  oled.print(freqDisplay);

  if (currentFrequency < 520 )
    bandMode = "LW  ";
  else
    bandMode = bandModeDesc[currentMode];

  oled.setCursor(0, 0);
  oled.print(bandMode);

  oled.setCursor(95, 0);
  oled.print(unit);
}

// Will be used by seekStationProgress
void showFrequencySeek(uint16_t freq)
{
  previousFrequency = currentFrequency = freq;
  showFrequency();
}

//    Show some basic information on display

void showStatus()
{
    sprintf(cadena,"St:%02d",currentStep);
    oled.setCursor(80, 1); oled.print(cadena);

    oled.setCursor(0, 3);

  if (currentMode == LSB || currentMode == USB)
  {
      sprintf(cadena,"BW:%sKHz",bandwidthSSB[bwIdxSSB]);
      oled.print(cadena);
      showBFO();
  }
  else if (currentMode == AM)
  {
      sprintf(cadena,"BW:%sKHz",bandwidthSSB[bwIdxAM]);
      oled.print(cadena);
  }

  // Show AGC Information
  si4735.getAutomaticGainControl();
  oled.setCursor(0, 1);
  oled.print((si4735.isAgcEnabled()) ? "AGC ON " : "AGC OFF");

  showRSSI();
  showVolume();
}

/* *******************************
   Shows RSSI status
*/
void showRSSI()
{
  oled.setCursor(97, 3);
  oled.print("S:"); oled.print(rssi);

  if ( currentMode == FM) {
    oled.setCursor(0, 3);
    oled.print((si4735.getCurrentPilot()) ? "ST  " : "MONO");
  }

}

/*
   Shows the volume level on LCD
*/
void showVolume()
{
  oled.setCursor(62, 3);
  oled.print("V:");
  oled.print(si4735.getCurrentVolume());
}

/*
   Shows the BFO current status.
   Must be called only on SSB mode (LSB or USB)
*/
void showBFO()
{
  if (currentBFO > 0)
    sprintf(cadena,"BFO:%+04dHz  st:%03d ",currentBFO,currentBFOStep);
  else
    sprintf(cadena,"BFO:%+04dHz  st:%03d ",currentBFO,currentBFOStep);
  oled.setCursor(0, 2);
  oled.print(cadena);
}

//   Goes to the next band (see Band table)

void bandUp()
{
  // save the current frequency for the band
  band[bandIdx].currentFreq = currentFrequency;
  band[bandIdx].currentStep = currentStep;

  if (bandIdx < lastBand)
    { bandIdx++; }
  else
  {   bandIdx = 0; }
  useBand();
}

/*
   Goes to the previous band (see Band table)
*/
void bandDown()
{
  // save the current frequency for the band
  band[bandIdx].currentFreq = currentFrequency;
  band[bandIdx].currentStep = currentStep;
  if (bandIdx > 0)
    { bandIdx--; }
  else
    { bandIdx = lastBand; }
  useBand();
}

/*
   This function loads the contents of the ssb_patch_content array into the CI (Si4735) and starts the radio on
   SSB mode.
*/
void loadSSB()
{
  oled.setCursor(0, 2);
  oled.print("  Switching to SSB  ");

  si4735.reset();
  si4735.queryLibraryId(); // Is it really necessary here?  Just powerDown() maigh work!
  si4735.patchPowerUp();
  delay(50);
  si4735.setI2CFastMode(); // Recommended
  // si4735.setI2CFastModeCustom(500000); // It is a test and may crash.
  si4735.downloadPatch(ssb_patch_content, size_content);
  si4735.setI2CStandardMode(); // goes back to default (100kHz)
 

  // delay(50);
  // Parameters
  // AUDIOBW - SSB Audio bandwidth; 0 = 1.2kHz (default); 1=2.2kHz; 2=3kHz; 3=4kHz; 4=500Hz; 5=1kHz;
  // SBCUTFLT SSB - side band cutoff filter for band passand low pass filter ( 0 or 1)
  // AVC_DIVIDER  - set 0 for SSB mode; set 3 for SYNC mode.
  // AVCEN - SSB Automatic Volume Control (AVC) enable; 0=disable; 1=enable (default).
  // SMUTESEL - SSB Soft-mute Based on RSSI or SNR (0 or 1).
  // DSP_AFCDIS - DSP AFC Disable or enable; 0=SYNC MODE, AFC enable; 1=SSB MODE, AFC disable.
  si4735.setSSBConfig(bwIdxSSB, 1, 0, 0, 0, 1);
  delay(25);
  ssbLoaded = true;
  oled.clear();
}

/*
   Switch the radio to current band.
   The bandIdx variable points to the current band. 
   This function change to the band referenced by bandIdx (see table band).
*/
void useBand()
{
  if (band[bandIdx].bandType == FM_BAND_TYPE)
  {
    currentMode = FM;
    si4735.setTuneFrequencyAntennaCapacitor(0);
    si4735.setFM(band[bandIdx].minimumFreq, band[bandIdx].maximumFreq, band[bandIdx].currentFreq, band[bandIdx].currentStep);
    bfoOn = ssbLoaded = false;
  }
  else
  {
    if (band[bandIdx].bandType == MW_BAND_TYPE || band[bandIdx].bandType == LW_BAND_TYPE)
      si4735.setTuneFrequencyAntennaCapacitor(0);
    else
      si4735.setTuneFrequencyAntennaCapacitor(1);

    if (ssbLoaded)
    {
      si4735.setSSB(band[bandIdx].minimumFreq, band[bandIdx].maximumFreq, band[bandIdx].currentFreq, band[bandIdx].currentStep, currentMode);
      si4735.setSSBAutomaticVolumeControl(1);
    }
    else
    {
      currentMode = AM;
      si4735.setAM(band[bandIdx].minimumFreq, band[bandIdx].maximumFreq, band[bandIdx].currentFreq, band[bandIdx].currentStep);
      si4735.setAutomaticGainControl(1, 0);
      bfoOn = false;
    }
    // Change this value (between 0 to 8); If 0, no Softmute;
    si4735.setSsbSoftMuteMaxAttenuation(4); // Work on AM and SSB -> 0 = Disable Soft Mute for SSB; 1 to 8 Softmute
    // Sets the seeking limits and space.
    si4735.setSeekAmLimits(band[bandIdx].minimumFreq, band[bandIdx].maximumFreq);               // Consider the range all defined current band
    si4735.setSeekAmSpacing((band[bandIdx].currentStep > 10) ? 10 : band[bandIdx].currentStep); // Max 10kHz for spacing

  }
  delay(100);
  currentFrequency = band[bandIdx].currentFreq;
  currentStep = band[bandIdx].currentStep;
  showStatus();
  showFrequency();
}


void loop()
{
  if (rDir != 0)
  {
    if (bfoOn)
      {
        currentBFO = (rDir == 1) ? (currentBFO + currentBFOStep) : (currentBFO - currentBFOStep);
        si4735.setSSBBfo(currentBFO);
        showBFO();
      }
    else
      {
        if (rDir == 1)   {  si4735.frequencyUp();    seekDirection = 1; }
           else       {  si4735.frequencyDown();  seekDirection = 0; }
        currentFrequency = si4735.getFrequency();
      }
    rDir = 0;
  }

  // Show the current frequency only if it has changed
  if (currentFrequency!=previousFrequency) { previousFrequency=currentFrequency; showFrequency(); }

  if(ti>10000)
        { 
          ti=0;
          if(info==1)
          {
            si4735.getCurrentReceivedSignalQuality();
              int aux = si4735.getCurrentRSSI();
              if (rssi != aux)
                  {
                    rssi = aux;
                    showRSSI();
                  }          
            showStatus(); 
          }
        }

  if(1==F_boton())  
      { 
        menuX();  
        showFrequency();
      }
  r.loop();  
  delayMicroseconds(100);
  ti++;
}



void showDirection(Rotary& r) 
{
  if(r.getDirection()==1) rDir=-1; else rDir=1;
}


////////////////////////////////SETA43////

void menuX()
{
  char e=0;
  modoMenu=1;
  printMenuX(e);
  while(1)
     {
         r.loop();
         delayMicroseconds(100);
        
         if(rDir!=0)
          {
            xStringAT(0,e," ");
            e+=rDir;
            if(e<1) e=0;
            if(e>=NMENUS) e=NMENUS-1;
            printMenuX(e);
            rDir=0;
          }
        if(1==F_boton())
          {
            if(e==0) 
              {
                  xclearDisplay(); modoMenu=0; //showBFO(); showStatus(); 
                  return; 
              }
            proceMenuX(e);
            printMenuX(e); 
          }
     }
}


void printMenuX(char op)
{
  xclearDisplay();
//  ModeFont=0;
  xStringAT(0,0,">");
  for(char z=0;z<4;z++)
    {
      if(op>=NMENUS) return;
      xStringAT(2,z,menus[op]);
      switch(op)
        {
          case 1:
            sprintf(cadena,"%02i",volume);
          break;
          case 2:
            sprintf(cadena,"%s",bandName[bandIdx]);
          break;
          case 3:
            sprintf(cadena,"%s",bandModeDesc[currentMode]);
          break;
          case 4:
            sprintf(cadena,"%s ",onOff[disableAgc]);
          break;
          case 5:
            sprintf(cadena,"%s ",bandwidthAM[bwIdxAM]);
          break;
          case 6:
            sprintf(cadena,"%s ",bandwidthSSB[bwIdxSSB]);
          break;
          case 7:
            sprintf(cadena,"%d ",currentStep);
          break;
          case 8:
            sprintf(cadena,"%s ",onOff[bfoOn]);
          break;
          case 9:
            sprintf(cadena,"%s ",onOff[info]);
          break;
          default:
            sprintf(cadena," ");
          break;
        }
       xStringAT(10,z,(char *)cadena);
       op++; 
    }
    
}



char proceMenuX(char op)
{
  xclearDisplay();
 // ModeFont=2;
  sprintf(cadena,"%s",menus[op]);  xStringAT(0,1,(char *)cadena);
    
  rDir=0; f2=1;
  switch(op)
    {
      case 1:
        proVolumen();
        si4735.setVolume(volume); 
      break;
      case 2:
        proBAND();
        useBand();
      break;
      case 3:
        if (currentMode != FM ) 
           {
            if (currentMode == AM)
              {
              // If you were in AM mode, it is necessary to load SSB patch (avery time)
              loadSSB();   currentMode = LSB;
              }
            else 
              if (currentMode == LSB)  {  currentMode = USB; }
              else 
                if (currentMode == USB)
                    {  currentMode = AM;  ssbLoaded = false;   bfoOn = false; }
            // Nothing to do if you are in FM mode
            band[bandIdx].currentFreq = currentFrequency;
            band[bandIdx].currentStep = currentStep;
            useBand();
        }
      break;
      case 4:
        disableAgc = !disableAgc;
        si4735.setAutomaticGainControl(disableAgc, 1);
        break;
      case 5:
        proBwAM();
      break;
      case 6:
        proBwSSB();
      break;   
      case 7:
        proSTEP();
      break;
      case 8:
        proBFO();
      break;  
      case 9:
        info=!info;
      break;    
    }
  xclearDisplay();
}



char menuValor2(char &var,char min,char max)
{
        r.loop();        
        delayMicroseconds(100);
        ti++;    
        if(rDir!=0)
          {
            var+=rDir;
            
            if(var<min) var=min;
            if(var>max) var=max;
            rDir=0; f2=1; ti=0;           
          }         
        if(1==F_boton()) {return(1); }
        if(ti>1000)
            {
              if(f2==1) {  f2=0;  return(2);  }
              ti=0;
            }     
        return(0);  
}

void proBFO()
{
  if (currentMode == LSB || currentMode == USB) 
      {
        bfoOn = !bfoOn;
        if (bfoOn)  showBFO();
      } 
      else 
      {
        si4735.seekStationProgress(showFrequencySeek, seekDirection);
        delay(MIN_ELAPSED_TIME); // waits a little more for releasing the button.
        currentFrequency = si4735.getFrequency();     
      }
  showStatus();
}


void proSTEP()
{
// This command should work only for SSB mode
        if (bfoOn && (currentMode == LSB || currentMode == USB))
        {
          currentBFOStep = (currentBFOStep == 25) ? 10 : 25;
          showBFO();
        }
        else
        {
          if (currentStep == 1)
            currentStep = 5;
          else if (currentStep == 5)
            currentStep = 10;
          else
            currentStep = 1;
          si4735.setFrequencyStep(currentStep);
          band[bandIdx].currentStep = currentStep;
 //         showStatus();
        }
}



void proBwSSB()
{
      bwIdxSSB++;
      if (bwIdxSSB > 5)
        bwIdxSSB = 0;
      si4735.setSSBAudioBandwidth(bwIdxSSB);
      // If audio bandwidth selected is about 2 kHz or below, it is recommended to set Sideband Cutoff Filter to 0.
      if (bwIdxSSB == 0 || bwIdxSSB == 4 || bwIdxSSB == 5)
        si4735.setSSBSidebandCutoffFilter(0);
      else
        si4735.setSSBSidebandCutoffFilter(1);
}


void proBwAM()
{
              bwIdxAM++;
              if (bwIdxAM > 6)  bwIdxAM = 0;
              si4735.setBandwidth(bwIdxAM, 1);
}


void proVolumen()
{ 
  while(1)
  {
    switch(menuValor2(volume,1,63))
      {
        case 1: //Salir
          return;  
          break;
        case 2:
          sprintf(cadena,"%i  ",volume);  
          xStringAT(10,1,(char *)cadena); 
          break;       
      }         
  }
}

void proBAND(void)
{
  while(1)
    {
      switch(menuValor2(bandIdx,0,lastBand))
        {
          case 1: //Salir
            return;  
            break;
          case 2:
            sprintf(cadena,"%s",bandName[bandIdx]);  
            xStringAT(1,3,(char *)cadena);
            break;       
        }         
    }
}




char F_boton(void)
{
  if (digitalRead(BUTTON_PIN) == 0) 
        {     
          delay(100);
          while(digitalRead(BUTTON_PIN) == 0) ;
          delay(100);
          //Serial.println("Boton");
          return(1);
        }
  return(0);
} 


///////////////////////////////////
void xclearDisplay(void) 
{
  oled.clear();
}

void xCaracterAT(int x,int y,char character)
{
  x*=6;
  y*=9;
    oled.setCursor(x,y);
    oled.print(character);
}

//Imprime cadena de caracteres
void xStringAT(unsigned char x,unsigned char y,  char *characters)
{
  while (*characters)
  {
    xCaracterAT(x,y, *characters++);
    x++;
  }

}
